package org.easystub;

import java.util.List;
import java.util.Arrays;

public class ResultReporter<T> {

    private final InvocationReporter invocationReporter;
    private final RecordedInvocationAnswerer answerer;
    private final ThrowableFactory throwableFactory = ThrowableFactory.instance();

    private static final List<Class<? extends Throwable>> EXCUSED_THROWABLES = Arrays.asList(RuntimeException.class,
                                                                                             Error.class);

    public ResultReporter(InvocationReporter invocationReporter, RecordedInvocationAnswerer answerer) {
        this.invocationReporter = invocationReporter;
        this.answerer = answerer;
    }

    public void withValue(T reply) {
        Invocation invocation = invocationReporter.lastInvocation();
        guardAgainstVoidReturnType(invocation);
        answerer.record(invocation, reply);
    }

    private void guardAgainstVoidReturnType(Invocation invocation) {
        if(Void.TYPE.equals(invocation.getMethodCalled().getReturnType())) {
            throw new UnsupportedOperationException("Cannot register reply on a void method");
        }
    }

    public void withException(Class<? extends Throwable> throwable) {
        Throwable instance = throwableFactory.create(throwable);
        withException(instance);
    }

    public void withException(Throwable throwable) {
        Invocation invocation = invocationReporter.lastInvocation();
        guardAgainstTooGenericThrowable(throwable, invocation);
        answerer.record(invocation, throwable);
    }

    private void guardAgainstTooGenericThrowable(Throwable throwable, Invocation invocation) {
        for(Class<? extends Throwable> excused : EXCUSED_THROWABLES) {
            if(excused.isAssignableFrom(throwable.getClass())) {
                return;
            }
        }
        
        Class<?>[] declaredThrowables = invocation.getMethodCalled().getExceptionTypes();

        for(Class<?> declared : declaredThrowables) {
            if(declared.isAssignableFrom(throwable.getClass())) {
                return;
            }
        }

        throw new UnsupportedOperationException("Cannot throw more generic exception than the one declared on the method.");
    }
}
